from django.db import migrations, transaction


def define_default_layout_roles(apps, schema_editor):

    # noinspection PyPep8Naming
    AnchorCategorySetting = apps.get_model('tour_guide', 'AnchorCategorySetting')
    DEFAULT_NUMBER_FORMAT = 'tour_guide:numbers'
    LETTERS_NUMBER_FORMAT = 'tour_guide:letters'
    DEFAULT_IDENTIFIER_FORMAT = 'tour_guide:generic'
    ANCHOR_IDENTIFIER_FORMAT = 'tour_guide:anchor_identifier'

    with transaction.atomic():
        figure_category = AnchorCategorySetting()
        figure_category.name = 'Figure'
        figure_category.identifier = 'figure'
        figure_category.label = 'Figure'
        figure_category.label_plural = 'Figures'
        figure_category.label_short = 'Fig'
        figure_category.number_format = DEFAULT_NUMBER_FORMAT
        figure_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        figure_category.save()

        image_category = AnchorCategorySetting()
        image_category.name = 'Image'
        image_category.identifier = 'image'
        image_category.label = 'Image'
        image_category.label_plural = 'Images'
        image_category.label_short = 'Img'
        image_category.number_format = DEFAULT_NUMBER_FORMAT
        image_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        image_category.save()

        table_category = AnchorCategorySetting()
        table_category.name = 'Table'
        table_category.identifier = 'table'
        table_category.label = 'Table'
        table_category.label_plural = 'Tables'
        table_category.label_short = 'T'
        table_category.number_format = DEFAULT_NUMBER_FORMAT
        table_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        table_category.save()

        equation_category = AnchorCategorySetting()
        equation_category.name = 'Equation'
        equation_category.identifier = 'equation'
        equation_category.label = 'Equation'
        equation_category.label_plural = 'Equations'
        equation_category.label_short = 'Eq'
        equation_category.number_format = DEFAULT_NUMBER_FORMAT
        equation_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        equation_category.save()

        listing_category = AnchorCategorySetting()
        listing_category.name = 'Code Listing'
        listing_category.identifier = 'listing'
        listing_category.label = 'Listing'
        listing_category.label_plural = 'Listings'
        listing_category.label_short = 'L'
        listing_category.number_format = DEFAULT_NUMBER_FORMAT
        listing_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        listing_category.save()

        section_category = AnchorCategorySetting()
        section_category.name = 'Section'
        section_category.identifier = 'section'
        section_category.label = 'Section'
        section_category.label_plural = 'Section'
        section_category.label_short = 'Sec'
        section_category.number_format = DEFAULT_NUMBER_FORMAT
        section_category.identifier_format = ANCHOR_IDENTIFIER_FORMAT
        section_category.save()

        slideshow_category = AnchorCategorySetting()
        slideshow_category.name = 'Slideshow'
        slideshow_category.identifier = 'slideshow'
        slideshow_category.label = 'Slideshow'
        slideshow_category.label_plural = 'Slideshow'
        slideshow_category.label_short = 'Slideshow'
        slideshow_category.number_format = DEFAULT_NUMBER_FORMAT
        slideshow_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        slideshow_category.save()

        reference_category = AnchorCategorySetting()
        reference_category.name = 'Reference'
        reference_category.identifier = 'ref'
        reference_category.label = ''
        reference_category.label_plural = ''
        reference_category.label_short = ''
        reference_category.number_format = LETTERS_NUMBER_FORMAT
        reference_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        reference_category.save()

        endnote_category = AnchorCategorySetting()
        endnote_category.name = 'Endnote'
        endnote_category.identifier = 'endnote'
        endnote_category.label = ''
        endnote_category.label_plural = ''
        endnote_category.label_short = ''
        endnote_category.link_css_class = 'endnote-link'
        endnote_category.label_css_class = 'endnote-label'
        endnote_category.number_format = DEFAULT_NUMBER_FORMAT
        endnote_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        endnote_category.save()

        aside_category = AnchorCategorySetting()
        aside_category.name = 'Aside'
        aside_category.identifier = 'aside'
        aside_category.label = ''
        aside_category.label_plural = ''
        aside_category.label_short = ''
        endnote_category.link_css_class = 'aside-link'
        endnote_category.label_css_class = 'aside-label'
        aside_category.number_format = DEFAULT_NUMBER_FORMAT
        aside_category.identifier_format = DEFAULT_IDENTIFIER_FORMAT
        aside_category.save()

        generic_category = AnchorCategorySetting()
        generic_category.name = 'Generic'
        generic_category.identifier = 'generic'
        generic_category.label = ''
        generic_category.label_plural = ''
        generic_category.label_short = ''
        generic_category.number_format = DEFAULT_NUMBER_FORMAT
        generic_category.identifier_format = ANCHOR_IDENTIFIER_FORMAT
        generic_category.save()


class Migration(migrations.Migration):

    dependencies = [
        ('tour_guide', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(define_default_layout_roles),
    ]



import re
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

__all__ = ['validate_identifier']


def validate_identifier(value):

    if re.search(r"(^[^A-Za-z0-9_]|\s)", value):

        raise ValidationError(
            _("Identifiers must start with a letter, a digit or underscore and cannot contain whitespace characters: {:}").format(value)
        )



import posixpath

from django.db.models import CharField, Q
from django.db.models.expressions import Exists, OuterRef
from django.db.models.functions import Length, Substr

from wagtail.models.sites import Site
from wagtail.search.queryset import SearchableQuerySetMixin
from wagtail.query import TreeQuerySet


class NavigationLinkQuerySet(SearchableQuerySetMixin, TreeQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _clone(self):
        clone = super()._clone()
        return clone

    def link_q(self, other):
        return Q(id=other.id)

    def link(self, other):
        """
        This filters the QuerySet so it only contains the specified link.
        """
        return self.filter(self.link_q(other))

    def not_link(self, other):
        """
        This filters the QuerySet so it doesn't contain the specified link.
        """
        return self.exclude(self.link_q(other))

    def first_common_ancestor(self, include_self=False, strict=False):
        """
        Find the first ancestor that all links in this queryset have in common.
        For example, consider a link hierarchy like::

            - Home/
                - Foo Event Index/
                    - Foo Event Page 1/
                    - Foo Event Page 2/
                - Bar Event Index/
                    - Bar Event Page 1/
                    - Bar Event Page 2/

        The common ancestors for some queries would be:

        .. code-block:: python

            >>> Link.objects\\
            ...     .first_common_ancestor()
            <Link: Home>
            >>> Link.objects\\
            ...     .filter(title__contains='Foo')\\
            ...     .first_common_ancestor()
            <Link: Foo Event Index>

        This method tries to be efficient, but if you have millions of pages
        scattered across your page tree, it will be slow.

        If `include_self` is True, the ancestor can be one of the pages in the
        queryset:

        .. code-block:: python

            >>> Link.objects\\
            ...     .filter(title__contains='Foo')\\
            ...     .first_common_ancestor()
            <Link: Foo Event Index>
            >>> Link.objects\\
            ...     .filter(title__exact='Bar Event Index')\\
            ...     .first_common_ancestor()
            <Link: Bar Event Index>

        A few invalid cases exist: when the queryset is empty, when the root
        Page is in the queryset and ``include_self`` is False, and when there
        are multiple page trees with no common root (a case Wagtail does not
        support). If ``strict`` is False (the default), then the first root
        node is returned in these cases. If ``strict`` is True, then a
        ``ObjectDoesNotExist`` is raised.
        """
        # An empty queryset has no ancestors. This is a problem
        if not self.exists():
            if strict:
                raise self.model.DoesNotExist("Can not find ancestor of empty queryset")
            return self.model.get_first_root_node()

        if include_self:
            # Get all the paths of the matched pages.
            paths = self.order_by().values_list("path", flat=True)
        else:
            # Find all the distinct parent paths of all matched pages.
            # The empty `.order_by()` ensures that `Page.path` is not also
            # selected to order the results, which makes `.distinct()` works.
            paths = (
                self.order_by()
                .annotate(
                    parent_path=Substr(
                        "path",
                        1,
                        Length("path") - self.model.steplen,
                        output_field=CharField(max_length=255),
                    )
                )
                .values_list("parent_path", flat=True)
                .distinct()
            )

        # This method works on anything, not just file system paths.
        common_parent_path = posixpath.commonprefix(paths)

        # That may have returned a path like (0001, 0002, 000), which is
        # missing some chars off the end. Fix this by trimming the path to a
        # multiple of `Page.steplen`
        extra_chars = len(common_parent_path) % self.model.steplen
        if extra_chars != 0:
            common_parent_path = common_parent_path[:-extra_chars]

        if common_parent_path == "":
            # This should only happen when there are multiple trees,
            # a situation that Wagtail does not support;
            # or when the root node itself is part of the queryset.
            if strict:
                raise self.model.DoesNotExist("No common ancestor found!")

            # Assuming the situation is the latter, just return the root node.
            # The root node is not its own ancestor, so this is technically
            # incorrect. If you want very correct operation, use `strict=True`
            # and receive an error.
            return self.model.get_first_root_node()

        # Assuming the database is in a consistent state, this page should
        # *always* exist. If your database is not in a consistent state, you've
        # got bigger problems.
        return self.model.objects.get(path=common_parent_path)

    def translation_of_q(self, link, inclusive):
        q = Q(translation_key=link.translation_key)

        if not inclusive:
            q &= ~Q(pk=link.pk)

        return q

    def translation_of(self, link, inclusive=False):
        """
        This filters the QuerySet to only contain links that are translations of the specified link.

        If inclusive is True, the link itself is returned.
        """
        return self.filter(self.translation_of_q(link, inclusive))

    def not_translation_of(self, link, inclusive=False):
        """
        This filters the QuerySet to only contain links that are not translations of the specified link.

        Note, this will include the link itself as the link is technically not a translation of itself.
        If inclusive is True, we consider the link to be a translation of itself so this excludes the link
        from the results.
        """
        return self.exclude(self.translation_of_q(link, inclusive))

    def annotate_site_root_state(self):
        """
        Performance optimisation for listing pages.
        Annotates each object with whether it is a root page of any site.
        Used by `is_site_root` method on `wagtailcore.models.Page`.
        """
        return self.annotate(
            _is_site_root=Exists(
                Site.objects.filter(
                    root_link__translation_key=OuterRef("translation_key")
                )
            )
        )




INLINE_ANCHOR_DEFINE_FEATURE_NAME = 'tour_guide.inline_anchor_define'
INLINE_ANCHOR_DEFINE_ENTITY_TYPE = 'tour_guide.INLINE_ANCHOR_DEFINE'

INLINE_ANCHOR_LINK_FEATURE_NAME = 'tour_guide.inline_anchor_link'
INLINE_ANCHOR_LINK_ENTITY_TYPE = 'tour_guide.INLINE_ANCHOR_LINK'

INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME = 'tour_guide.inline_anchor_labelled_link'
INLINE_ANCHOR_LABELLED_LINK_ENTITY_TYPE = 'tour_guide.INLINE_ANCHOR_LABELLED_LINK'

feature_names = [INLINE_ANCHOR_DEFINE_FEATURE_NAME,
                 INLINE_ANCHOR_LINK_FEATURE_NAME,
                 INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME
                 ]



from urllib.parse import urlparse

from model_porter.config import ModelPorterConfig
from model_porter.model_porter import WagtailFunctions

from media_catalogue.model_porter import media_chooser_value

from .blocks import RichLink, RichLinkBlock, RichLinkBlockValue

DEFAULT_RICH_LINK_BLOCK = RichLinkBlock()


def define_rich_link(*, specifier, context):

    link = RichLink()

    if not isinstance(specifier, str):

        # Try dict
        link.rich_text = specifier.get("rich_text", None)
        link.display_context = specifier.get("display_context", None)
        link.relationships = specifier.get("relationships", [])
        media_item = specifier.get("media_item", None)

        if media_item:
            media_item, _ = media_chooser_value(items=[media_item], context=context)[0]

        if media_item:
            link.media_item = media_item

        link.external_url = specifier.get("external_url", None)
        link.page_fragment = specifier.get("page_fragment", None)

        page = specifier.get("page", None)
        parts = urlparse(page)

        if parts.scheme:
            page = context.get_instance(page)
        else:
            page = WagtailFunctions.locate_page_for_path(page)

        link.page = page

    elif isinstance(specifier, str):

        # <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

        parts = urlparse(specifier)

        if parts.scheme or parts.netloc or not parts.path:
            link.external_url = specifier
        else:
            page = WagtailFunctions.locate_page_for_path(parts.path)

            if page is not None:
                link.page = page
                link.page_fragment = parts.fragment
            else:
                link.external_url = specifier

    return link


def define_rich_link_block(*, specifier, context):
    link = define_rich_link(specifier=specifier, context=context)
    return RichLinkBlockValue.from_rich_link(DEFAULT_RICH_LINK_BLOCK, link)


def create_or_update_navigation_style(*, model=None, identifier, context):
    # model argument is not used which allows this method to be used for
    # regular attributes instead of only as a create method.

    if model is None:
        from .models import NavigationStyle
        model = NavigationStyle

    try:
        instance = model.objects.get(identifier=identifier)
    except model.DoesNotExist:
        instance = model()

    return instance


class TourGuideConfig(ModelPorterConfig):

    def __init__(self, app_label, module):
        super(TourGuideConfig, self).__init__(app_label, module)
        self.register_function_action(define_rich_link, context_argument='context')
        self.register_function_action(define_rich_link_block, context_argument='context')
        self.register_function_action(create_or_update_navigation_style, context_argument='context')



# Generated by Django 4.1.7 on 2023-05-11 11:51

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import model_porter.support_mixin
import tour_guide.blocks
import tour_guide.model_fields
import tour_guide.validators
import wagtail_block_model_field.fields
import wagtail_dynamic_choice.model_fields
import wagtail_switch_block.blocks


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0078_referenceindex'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AnchorCategorySetting',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('identifier', models.CharField(max_length=128, unique=True, validators=[tour_guide.validators.validate_identifier])),
                ('label', models.CharField(blank=True, default='Figure', max_length=128)),
                ('label_plural', models.CharField(blank=True, default='Figures', max_length=128)),
                ('label_short', models.CharField(blank=True, default='Fig', max_length=128)),
                ('link_css_class', models.CharField(blank=True, default='', max_length=128)),
                ('label_css_class', models.CharField(blank=True, default='', max_length=128)),
                ('number_format', wagtail_dynamic_choice.model_fields.DynamicChoiceField(choices_function_name='tour_guide.anchors.number_format_choices', default='numbers', max_length=128)),
                ('identifier_format', wagtail_dynamic_choice.model_fields.DynamicChoiceField(choices_function_name='tour_guide.anchors.identifier_format_choices', default='generic', max_length=128)),
            ],
            options={
                'verbose_name': 'Anchor Category',
                'verbose_name_plural': 'Anchor Categories',
            },
        ),
        migrations.CreateModel(
            name='NavigationStyle',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(max_length=128, validators=[django.core.validators.RegexValidator('^[A-Za-z_][A-Za-z_0-9]*$', code='invalid_identifier', message='A valid identifier starts with an alphanumeric letter or underscore and contains only alphanumeric letters, underscores or digits.')])),
                ('name', models.CharField(max_length=128)),
                ('builder', wagtail_block_model_field.fields.BlockModelField(tour_guide.blocks.NavigationBuilderBlock(), wagtail_switch_block.blocks.SwitchValue)),
                ('presenter', wagtail_block_model_field.fields.BlockModelField(tour_guide.blocks.NavigationPresenterBlock(), wagtail_switch_block.blocks.SwitchValue)),
                ('created_at', models.DateTimeField(default=None, editable=False, verbose_name='created at')),
                ('created_by_user', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='created by user')),
            ],
            options={
                'verbose_name': 'Navigation Style',
                'verbose_name_plural': 'Navigation Styles',
            },
        ),
        migrations.CreateModel(
            name='NavigationTarget',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('rich_link', wagtail_block_model_field.fields.BlockModelField(tour_guide.blocks.RichLinkBlock(label='Rich Link'), tour_guide.blocks.RichLinkBlockValue)),
                ('navigation_categories', tour_guide.model_fields.NavigationCategoriesField(blank=True, choices_function_name='tour_guide.categories.navigation_category_choices', max_length=128, verbose_name='Navigation Categories')),
            ],
            options={
                'verbose_name': 'Navigation Target',
                'verbose_name_plural': 'Navigation Targets',
            },
            bases=(model_porter.support_mixin.ModelPorterSupportMixin, 'wagtailcore.page'),
        ),
        migrations.CreateModel(
            name='NavigationStyleAlias',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(default='', max_length=128, validators=[django.core.validators.RegexValidator('^[A-Za-z_][A-Za-z_0-9]*$', code='invalid_identifier', message='A valid identifier starts with an alphanumeric letter or underscore and contains only alphanumeric letters, underscores or digits.')])),
                ('style', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='aliases', to='tour_guide.navigationstyle')),
            ],
            options={
                'verbose_name': 'Navigation Alias',
                'verbose_name_plural': 'Navigation Aliases',
            },
        ),
        migrations.AddConstraint(
            model_name='anchorcategorysetting',
            constraint=models.UniqueConstraint(fields=('identifier',), name='unique_tour_guide_anchorcategorysetting.identifier'),
        ),
        migrations.AddConstraint(
            model_name='anchorcategorysetting',
            constraint=models.UniqueConstraint(fields=('name',), name='unique_tour_guide_anchorcategorysetting.name'),
        ),
        migrations.AddConstraint(
            model_name='navigationstylealias',
            constraint=models.UniqueConstraint(fields=('identifier',), name='unique_tour_guide_navigationstylealias.identifier'),
        ),
        migrations.AddConstraint(
            model_name='navigationstyle',
            constraint=models.UniqueConstraint(fields=('identifier',), name='unique_tour_guide_navigationstyle.identifier'),
        ),
    ]



# -*- coding: utf-8 -*-

import sys
from django.conf import settings
from django.db import DEFAULT_DB_ALIAS
from django.apps import AppConfig
from django.urls import reverse
from django.utils.translation import gettext_lazy as _


def is_running_without_database():
    engine = settings.DATABASES[DEFAULT_DB_ALIAS]['ENGINE']
    return engine == 'django.db.backends.dummy'


class TourGuideConfig(AppConfig):
    name = 'tour_guide'
    label = 'tour_guide'
    verbose_name = _("Tour Guide")
    default_auto_field = 'django.db.models.BigAutoField'

    def import_models(self):

        from django_auxiliaries.variable_scope import register_variable_scope, Copy, EnvironmentVariable
        from .anchors import cast_to_anchor_registry

        register_variable_scope(self.label,
                                navigation_index=0,
                                anchor_inventory=Copy({}),
                                anchor_registry=EnvironmentVariable('page', None, copy_default=False,
                                                                    modifier=cast_to_anchor_registry))

        super().import_models()

    def ready(self):

        if is_running_without_database() or "makemigrations" in sys.argv or "migrate" in sys.argv:
            return


def get_app_label():
    return TourGuideConfig.label


def reverse_app_url(identifier):
    return reverse(f'{TourGuideConfig.label}:{identifier}')





from django import template
from django.templatetags.static import static

from wagtail.models import Site

from django_auxiliaries.tags import register_simple_block_tag

from ..models import lookup_navigation_style
from ..anchors import current_anchor_registry_and_inventory


from ..apps import get_app_label


APP_LABEL = get_app_label()

register = template.Library()


@register.simple_tag(takes_context=True)
def get_site_root(context):
    return Site.find_for_request(context['request']).root_page


@register.simple_tag(takes_context=True)
def link_url(context, *, link, request=None):

    if not request:
        request = context.get('request', None)

    return link.determine_url(request)


@register.simple_tag(takes_context=True, name="navigation")
def parse_navigation_tag(context, *, style, request=None, page=None, render_location=None, **kwargs):

    navigation_style = lookup_navigation_style(style)

    if navigation_style is None:
        return ''

    if not request:
        request = context.get('request', None)

    if page is None:
        page = context.get('page', None)

    return navigation_style.render(request, page, render_location, context=context.flatten())


@register.simple_tag(name="define_anchor")
def parse_define_anchor_tag(category, identifier, *, level=0, label_prefix=''):

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    category = str(category)
    identifier = str(identifier)
    level = int(level)
    label_prefix = str(label_prefix)

    result = anchor_registry.define_anchor(inventory, category, identifier, level, label_prefix)
    return result


@register_simple_block_tag(register, name="link_anchor", autoescape=False)
def parse_link_anchor_tag(context, nodelist, category, identifier):

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    inner = nodelist.render(context)

    category = str(category)
    identifier = str(identifier)

    result = anchor_registry.link_anchor(inventory, category, identifier, inner)
    return result


@register.simple_tag(name="anchor_url")
def parse_anchor_url_tag(category, identifier):

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    category = str(category)
    identifier = str(identifier)

    result = anchor_registry.url_for_anchor(inventory, category, identifier)
    return result


@register.simple_tag(name="unique_anchor_id")
def parse_unique_anchor_id_tag():

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    result = anchor_registry.generate_unique_id(inventory)
    return result


@register.simple_tag(name="count_category")
def parse_count_category_tag(category):

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    category = str(category)

    result = anchor_registry.count_category(inventory, category)
    return result


@register.simple_tag(name="label_anchor")
def parse_label_anchor_tag(category, identifier, *, label_format=None):

    anchor_registry, inventory = current_anchor_registry_and_inventory()

    if not anchor_registry or (not inventory and not isinstance(inventory, dict)):
        return ""

    category = str(category)
    identifier = str(identifier)

    result = anchor_registry.label_anchor(inventory, category, identifier, label_format)
    return result


def gather_style_sheets():

    urls = [static(APP_LABEL + '/css/dropdown.css'),
            static(APP_LABEL + '/css/dropdown_style.css'),
            static(APP_LABEL + '/css/breadcrumbs.css'),
            static(APP_LABEL + '/css/breadcrumbs_style.css'),
            static(APP_LABEL + '/css/site_map.css'),
            static(APP_LABEL + '/css/site_map_style.css')]

    return urls


def gather_scripts():

    urls = [static(APP_LABEL + '/js/pointer-navigation.js')]

    return urls


SUPPORT_TEMPLATE_SETTING = APP_LABEL + '/tags/support.html'


@register.inclusion_tag(SUPPORT_TEMPLATE_SETTING, name="tour_guide_support")
def tour_guide_support_tag(*, container_element, is_admin_page=False):

    result = {
        'container_element': container_element,
        'is_admin_page': is_admin_page,
        'stylesheets': gather_style_sheets(),
        'scripts': gather_scripts()
    }

    return result




from wagtail.permission_policies.collections import CollectionOwnershipPermissionPolicy

from .apps import get_app_label

APP_LABEL = get_app_label()

navigation_style_permission_policy = CollectionOwnershipPermissionPolicy(
    APP_LABEL + ".navigationstyle",
    owner_field_name='created_by_user'
)




class AnchorMiddleware(object):

    """

    AnchorMiddleware no longer needs to be installed as middleware, but instead is used by the AnchorRegistry
    page mixin inside its serve() method.


    """

    def __init__(self, get_response):
        self.current_anchor_registry_and_inventory = None
        self.get_response = get_response

    def __call__(self, request):

        response = self.get_response(request)

        # Deactivated because it introduces whitespace which might mess up the appearance of the output
        # if settings.DEBUG and response.get('Content-Type', ";").split(';', 1)[0] == 'text/html':
        #    response.content = BeautifulSoup(response.content, "html5lib").prettify()

        return response

    def process_template_response(self, request, response):

        if self.current_anchor_registry_and_inventory is None:
            from .anchors import current_anchor_registry_and_inventory
            self.current_anchor_registry_and_inventory = current_anchor_registry_and_inventory

        if response.get('Content-Type', ";").split(';', 1)[0] != 'text/html':
            return response

        if response.context_data is None:
            return response

        registry, inventory = self.current_anchor_registry_and_inventory()

        if not registry:
            return response

        registry.reset()

        content = str(response.rendered_content)

        tokens = self.parse_tokens(registry, content)
        self.order_tokens(registry, inventory, tokens)
        processed_content = self.apply_tokens(registry, inventory, tokens, content)
        response.content = processed_content.encode(encoding=response.charset)

        return response

    # noinspection PyMethodMayBeStatic
    def order_tokens(self, registry, template_context, tokens):

        for token in tokens:
            if token.directive != 'define':
                continue

            registry.order_anchor(template_context, token.category, token.identifier, token.level)

    # noinspection PyMethodMayBeStatic
    def apply_tokens(self, registry, template_context, tokens, text):

        result = ''
        pos = 0

        for token in tokens:

            if pos < token.start_offset:
                result += text[pos:token.start_offset]

            token_result = registry.resolve_token(template_context,
                                                  token.directive, token.category, token.identifier)

            result += token_result
            pos = token.end_offset

        if pos < len(text):
            result += text[pos:]

        return result

    # noinspection PyMethodMayBeStatic
    def parse_tokens(self, registry, text):

        tokens = []
        pos = 0

        while pos < len(text):

            token = registry.match_next_token(text, pos)

            if token is None:
                break

            tokens.append(token)
            pos = token.end_offset

        return tokens




from wagtail_dynamic_choice.model_fields import DynamicMultipleChoiceField

from .apps import get_app_label

__all__ = ['NavigationCategoriesField']

APP_LABEL = get_app_label()


class NavigationCategoriesField(DynamicMultipleChoiceField):

    def __init__(self, default=None, blank=True, **kwargs):

        self.default_argument = default

        if default is None:
            from .categories import navigation_category_choices

            choices = navigation_category_choices()

            def default():
                if choices:
                    return [choices[0][0]]
                else:
                    return []

        if 'max_length' not in kwargs:
            kwargs['max_length'] = 128

        if 'choices_function_name' not in kwargs:
            kwargs['choices_function_name'] = APP_LABEL + ".categories.navigation_category_choices"

        super().__init__(default=default, blank=blank, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        kwargs.pop('default', None)

        if self.default_argument:
            kwargs['default'] = self.default_argument

        return name, path, args, kwargs



import os

from django.http import Http404
import django.db.models as django_models

from django.core.files.storage import default_storage
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from django.utils import timezone


from wagtail.models import Page
from wagtail.snippets.models import register_snippet

# Panels

from wagtail.admin.panels import FieldPanel, MultiFieldPanel, FieldRowPanel

from django_auxiliaries.validators import python_identifier_validator
from wagtail_block_model_field.fields import BlockModelField
from wagtail_dynamic_choice.model_fields import DynamicChoiceField

from model_porter.support_mixin import ModelPorterSupportMixin

from .model_porter import define_rich_link_block

from .anchors import number_format_choices
from .blocks import RichLinkBlock, RichLinkBlockValue, NavigationBuilderBlock, NavigationBuilderBlockValue, \
                    NavigationPresenterBlock, NavigationPresenterBlockValue


from .decorators import chooses_navigation_categories

from .validators import validate_identifier
from .apps import get_app_label

__all__ = ['NavigationTarget', 'NavigationStyle', 'NavigationStyleAlias', 'lookup_navigation_style',
           'AnchorCategorySetting']

APP_LABEL = get_app_label()


def remove_title_field(content_panels):
    result = [p for p in content_panels if p.field_name != 'title']
    return result


@chooses_navigation_categories
class NavigationTarget(ModelPorterSupportMixin, Page):

    class Meta:
        verbose_name = 'Navigation Target'
        verbose_name_plural = 'Navigation Targets'

    rich_link = BlockModelField(
        RichLinkBlock(label="Rich Link"),
        RichLinkBlockValue
    )

    content_panels = Page.content_panels + [
        FieldPanel('rich_link', heading="Rich Link")
    ]

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def route(self, request, path_components):

        if not path_components:
            # Navigation placeholders are non-routeable! # noqa
            raise Http404

        return super().route(request, path_components)

    def get_url_parts(self, request=None):
        return None

    def from_repository(self, value, context):

        rich_link = value.get("rich_link", None)

        if rich_link:
            rich_link = define_rich_link_block(specifier=rich_link, context=context)
            value["rich_link"] = rich_link

        return value


def get_schema_upload_path(self, filename):

    # if not default_storage.exists(filename):
    filename = default_storage.get_valid_name(filename)

    # noinspection PyUnresolvedReferences
    filename = filename.encode('ascii', errors='replace').decode('ascii')

    # noinspection PyUnresolvedReferences
    path = os.path.join(self.path, filename)

    if len(path) >= 95:
        chars_to_trim = len(path) - 94
        prefix, extension = os.path.splitext(filename)
        filename = prefix[:-chars_to_trim] + extension
        # noinspection PyUnresolvedReferences
        path = os.path.join(self.path, filename)

    return path


@register_snippet
class NavigationStyle(django_models.Model):

    # base_form_class = NavigationStyleForm

    class Meta:
        verbose_name = 'Navigation Style'
        verbose_name_plural = 'Navigation Styles'
        constraints = [
            django_models.UniqueConstraint(fields=['identifier'],
                                           name='unique_%(app_label)s_%(class)s.identifier')
        ]

    identifier = django_models.CharField(max_length=128, validators=[python_identifier_validator])
    name = django_models.CharField(max_length=128)

    builder = BlockModelField(NavigationBuilderBlock(), value_class=NavigationBuilderBlockValue)
    presenter = BlockModelField(NavigationPresenterBlock(), value_class=NavigationPresenterBlockValue)

    created_at = django_models.DateTimeField(
        verbose_name=_('created at'),
        default=None,
        editable=False)

    created_by_user = django_models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('created by user'),
        null=True,
        blank=True,
        editable=False,
        on_delete=django_models.SET_NULL
    )

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)   # noinspection PyUnresolvedReferences

        if not self.id: # noqa

            self.id = None
            self.created_at = timezone.now()

    def __str__(self):
        return '{}: {}'.format(self.identifier, self.name)

    def render(self, request, page, render_location, context=None):

        builder_block = self.__class__.builder.field.block_def # noqa
        presenter_block = self.__class__.presenter.field.block_def # noqa

        navigation = builder_block.build_navigation(self.builder, page, request)

        result = presenter_block.render_navigation(self.presenter, navigation, render_location, request, context)
        return result

    panels = [

        MultiFieldPanel([

            FieldPanel('identifier'),
            FieldPanel('name'),

        ], "Description"),

        FieldPanel('builder'),
        FieldPanel('presenter')
    ]


def navigation_style_choices():
    return [(style.identifier, style.name) for style in NavigationStyle.objects.all()]


@register_snippet
class NavigationStyleAlias(django_models.Model):

    class Meta:
        verbose_name = "Navigation Alias"
        verbose_name_plural = "Navigation Aliases"

        constraints = [
            django_models.UniqueConstraint(fields=['identifier'], name='unique_%(app_label)s_%(class)s.identifier')
        ]

    identifier = django_models.CharField(max_length=128, default='', validators=[python_identifier_validator])
    style = django_models.ForeignKey('NavigationStyle', related_name="aliases", on_delete=django_models.SET_NULL,
                                     blank=True, null=True)

    panels = [
        FieldRowPanel([
            FieldPanel('identifier'),
            FieldPanel('style')
        ])
    ]

    def __str__(self):
        return "{} -> {}".format(self.identifier, self.style.name if self.style else "[undefined]")


def lookup_navigation_style(identifier):

    navigation_style = None

    try:
        alias = NavigationStyleAlias.objects.get(identifier=identifier)
        navigation_style = alias.style

    except NavigationStyleAlias.DoesNotExist:
        pass

    if navigation_style is None:
        try:
            navigation_style = NavigationStyle.objects.get(identifier=identifier)
        except NavigationStyle.DoesNotExist:
            pass

    return navigation_style


@register_snippet
class AnchorCategorySetting(django_models.Model):

    class Meta:
        verbose_name = 'Anchor Category'
        verbose_name_plural = 'Anchor Categories'
        constraints = [
            django_models.UniqueConstraint(fields=['identifier'], name='unique_%(app_label)s_%(class)s.identifier'),
            django_models.UniqueConstraint(fields=['name'], name='unique_%(app_label)s_%(class)s.name')
        ]

    name = django_models.CharField(max_length=128, unique=True, blank=False, null=False)
    identifier = django_models.CharField(max_length=128, unique=True, blank=False, null=False, validators=[validate_identifier])
    label = django_models.CharField(max_length=128, blank=True, default='Figure')
    label_plural = django_models.CharField(max_length=128, blank=True, default='Figures')
    label_short = django_models.CharField(max_length=128, blank=True, default='Fig')

    link_css_class = django_models.CharField(max_length=128, blank=True, default='')
    label_css_class = django_models.CharField(max_length=128, blank=True, default='')

    number_format = DynamicChoiceField(max_length=128,
                                       choices_function_name=APP_LABEL + ".anchors.number_format_choices",
                                       default='numbers')

    identifier_format = DynamicChoiceField(max_length=128,
                                           choices_function_name=APP_LABEL + ".anchors.identifier_format_choices",
                                           default='generic')

    panels = [
        FieldPanel('name'),
        FieldPanel('identifier'),
        FieldPanel('label'),
        FieldPanel('label_plural'),
        FieldPanel('label_short'),
        FieldPanel('link_css_class'),
        FieldPanel('label_css_class'),
        FieldPanel('number_format'),
        FieldPanel('identifier_format')
    ]

    def __str__(self):

        number_format = [number_format for identifier, number_format in number_format_choices()
                         if identifier == self.number_format]

        if not number_format:
            number_format = 'none'
        else:
            number_format = number_format[0]
            number_format = number_format.format(singular=self.label, plural=self.label_plural)

        result = self.name + "/" + self.identifier + ": " + number_format # noqa
        return result




from django.utils.safestring import mark_safe

from wagtail import hooks

from .apps import get_app_label
from .anchor_features import feature_registrations as anchor_features
from .anchors import anchor_category_choices_as_json

APP_LABEL = get_app_label()

features_to_register = anchor_features

for register_feature in features_to_register:
    hooks.register('register_rich_text_features', fn=register_feature)


@hooks.register("insert_editor_js")
def editor_js():

    result = mark_safe('<script type="text/{}_anchor_categories">{}</script>'.format(
                APP_LABEL, anchor_category_choices_as_json()))
    return result



# -*- coding: utf-8 -*-

from collections import Sequence

from django.utils.translation import get_language_from_request

from wagtail.models import Site

from .apps import get_app_label

__all__ = ['root_page_for_request', 'select_page_children']

APP_LABEL = get_app_label()


def root_page_for_request(request):

    site = Site.find_for_request(request)

    if site is None:
        site = Site.objects.select_related("root_page").get(is_default_site=True)

    language_code = get_language_from_request(request, check_path=True)

    page = site.root_page

    for translation in page.get_translations().live():

        if translation.locale.language_code == language_code:
            page = translation
            break

    return page


def get_page_categories(page):

    result = {"menu"} if page.show_in_menus else set()

    if hasattr(page, "navigation_categories"):
        navigation_categories = page.navigation_categories

        if isinstance(navigation_categories, Sequence) and not isinstance(navigation_categories, str):
            for category in navigation_categories:
                result.add(category)

    return result


def select_page_children(page, navigation_categories):
    result = [child for child in page.get_children().live()
              if not get_page_categories(child.specific).isdisjoint(navigation_categories)]
    return result



# -*- coding: utf-8 -*-

from collections import namedtuple

from django import forms
from django.utils.functional import cached_property

from django.utils.safestring import mark_safe
from django.template.loader import render_to_string

from wagtail import blocks
from wagtail.models import Page

from wagtail.telepath import register
from wagtail.blocks.struct_block import StructBlockAdapter
from wagtail.admin.staticfiles import versioned_static


from django_auxiliaries.variable_scope import load_variable_scope
from csskit.blocks import ComplexCSSUnitValueBlock

from wagtail_switch_block import DynamicSwitchBlock, SwitchBlock, SwitchValue
from wagtail_switch_block.block_registry import BlockRegistry

from wagtail_dynamic_choice.blocks import AlternateSnippetChooserBlock, DynamicMultipleChoiceBlock
from wagtail_richer_text.blocks import RichTextBlock
from wagtail_preference_blocks.targets import PreferenceTarget

from model_porter.support_mixin import ModelPorterSupportMixin
from model_porter.model_porter import WagtailFunctions

from django_auxiliaries.templatetags.django_auxiliaries_tags import tagged_static

from .apps import get_app_label
from .links import DISPLAY_CONTEXTS, LINK_RELATIONSHIPS, RichLink
from .navigation import NavigationNode
from .utilities import root_page_for_request, select_page_children

APP_LABEL = get_app_label()

__all__ = ["RichLinkBlock", "RichLinkBlockValue",
           "NavigationBlock",
           "NavigationBuilderBlock", "NavigationBuilderBlockValue", "NavigationBuilderMixin",
           "register_navigation_builder_block", # noqa
           "navigation_builder_block_choices", # noqa
           "NavigationPresenterBlock", "NavigationPresenterBlockValue", "NavigationPresenterMixin",
           "register_navigation_presenter_block", # noqa
           "navigation_presenter_block_choices", # noqa
           "PagePathBuilderBlock", "PageTreeBuilderBlock",
           "InlineDefinitionBlock", "BeginNavigationLevelBlock", "EndNavigationLevelBlock",
           "TreeTemplatesBlock", "DropdownPresenterBlock", "BreadcrumbsPresenterBlock", "SiteMapPresenterBlock",
           "SITE_NAVIGATION_PREFERENCE_TARGET",
           "BREADCRUMBS_PREFERENCE_TARGET",
           "SITE_MAP_PREFERENCE_TARGET"]


class NavigationBlock(blocks.StructBlock):

    class Meta:
        icon = 'link'
        template = APP_LABEL + '/blocks/navigation_block.html'
        default = {}

    identifier = SwitchBlock(local_blocks=[
        ('alias', AlternateSnippetChooserBlock(APP_LABEL + ".navigationstylealias", use_identifier_as_value=True)),
        ('style', AlternateSnippetChooserBlock(APP_LABEL + ".navigationstyle", use_identifier_as_value=True))
    ])

    # noinspection PyMethodMayBeStatic
    def lookup_navigation_style(self, value):

        from .models import lookup_navigation_style
        return lookup_navigation_style(value['identifier'])

    def deconstruct(self):
        path, args, kwargs = blocks.Block.deconstruct(self)
        return path, args, kwargs


class RichLinkBlockValue(blocks.StructValue):

    @classmethod
    def from_rich_link(cls, block, rich_link):

        items = {
            'rich_text': rich_link.rich_text,
            'display_context': rich_link.display_context,
            'relationships': rich_link.relationships,
            'page': rich_link.page.pk if rich_link.page else None,
            'page_fragment': rich_link.page_fragment,
            'external_url': rich_link.external_url
        }

        result = block.to_python(items)
        return result

    def as_rich_link(self):

        rich_link = RichLink()

        for key, value in self.items():

            if hasattr(rich_link, key):
                setattr(rich_link, key, value)

        return rich_link

    def as_navigation_node(self):

        node = NavigationNode()

        for key, value in self.items():

            if hasattr(node, key):
                setattr(node, key, value)

        return node


class RichLinkBlock(ModelPorterSupportMixin, blocks.StructBlock):
    class Meta:
        icon = 'link'
        template = get_app_label() + '/blocks/rich_link_block.html'
        value_class = RichLinkBlockValue
        form_classname = 'rich-link-block struct-block'
        default = {}
        user_editable_display_context = True
        user_editable_relationships = True

    rich_text = RichTextBlock(label="Text", required=False, editor=APP_LABEL + ".richlinktext")

    page = blocks.PageChooserBlock(label="Page", required=False)

    page_fragment = blocks.CharBlock(label="Page Fragment", required=False,
                                     help_text="Optional identifier of element on page")

    external_url = blocks.URLBlock(label="Destination External URL", required=False)

    display_context = blocks.ChoiceBlock(label="Display Context", default="self", required=False,
                                         choices=DISPLAY_CONTEXTS)

    relationships = blocks.MultipleChoiceBlock(label="Link Relationships", default=[], required=False,
                                               choices=LINK_RELATIONSHIPS)

    def deconstruct(self):
        path, args, kwargs = blocks.Block.deconstruct(self)
        return path, args, kwargs

    def get_form_state(self, value):

        value = self.meta.value_class(self, value)

        if not self.meta.user_editable_display_context:
            value.pop('display_context', None)

        if not self.meta.user_editable_relationships:
            value.pop('relationships', None)

        return super().get_form_state(value)

    def clean(self, value):

        if not self.meta.user_editable_display_context:
            value['display_context'] = self.child_blocks['display_context'].get_default()

        if not self.meta.user_editable_relationships:
            value['relationships'] = self.child_blocks['relationships'].get_default()

        return super().clean(value)

    # noinspection PyMethodMayBeStatic
    def from_repository(self, value, context):

        page = value['page']
        page = WagtailFunctions.locate_page_for_path(page)
        value['page'] = page.id

        return value


class RichLinkBlockAdapter(StructBlockAdapter):

    def js_args(self, block):

        svd_child_blocks = block.child_blocks
        block.child_blocks = dict(svd_child_blocks)

        if not block.meta.user_editable_display_context:
            block.child_blocks.pop('display_context', None)

        if not block.meta.user_editable_relationships:
            block.child_blocks.pop('relationships', None)

        result = super().js_args(block)

        block.child_blocks = svd_child_blocks

        return result

    @cached_property
    def media(self):

        return super().media + forms.Media(
            css={
                "all": [tagged_static(APP_LABEL + "/css/tour_guide.css"),
                        tagged_static(APP_LABEL + "/css/tour_guide_extra.css")
                       ]
            },
            js=[
                versioned_static("wagtailadmin/js/draftail.js"),
                tagged_static(APP_LABEL + "/js/tour_guide.js")
            ]
        )


register(RichLinkBlockAdapter(), RichLinkBlock)


class NavigationBuilderBlockRegistry(BlockRegistry):

    # noinspection PyMethodMayBeStatic
    def should_register_block(self, app_label, local_identifier, block_type, block_args, block_kwargs, **kwargs):

        if not issubclass(block_type, NavigationBuilderMixin):
            raise RuntimeError("Registered block type must be a subclass of NavigationBuilderMixin")

        return True

    # noinspection PyMethodMayBeStatic
    def should_include_entry_for_container_block(self, identifier, entry, container_block):
        return True

    # noinspection PyMethodMayBeStatic
    def instantiate_block(self, identifier, entry, container_block):

        block_kwargs = dict(entry.block_kwargs)

        block = entry.block_type(*entry.block_args, **block_kwargs)
        block.set_name(identifier)

        return block


NAVIGATION_BUILDER_BLOCK_REGISTRY = NavigationBuilderBlockRegistry()
NAVIGATION_BUILDER_BLOCK_REGISTRY.define_procedures_in_caller_module("navigation_builder")


class NavigationBuilderMixin:

    def build_navigation(self, configuration, page, request, context=None):
        return None


NavigationBuilderBlockValue = SwitchValue


class NavigationBuilderBlock(DynamicSwitchBlock):

    class Meta:
        builder_blocks_function_name = APP_LABEL + ".blocks.navigation_builder_block_choices"
        choice_label = "Select Navigation Builder"

    def __init__(self, *args, **kwargs):

        builder_blocks_function_name = kwargs.pop("builder_blocks_function_name",
                                                  self._meta_class.builder_blocks_function_name) # noqa

        super().__init__(*args,
                         builder_blocks_function_name=builder_blocks_function_name,
                         child_blocks_function_name=builder_blocks_function_name,
                         **kwargs)

    def deconstruct(self):
        path, args, kwargs = blocks.Block.deconstruct(self)
        return path, args, kwargs

    def build_navigation(self, configuration, page, request, context=None):

        builder_block = self.child_blocks[configuration.type]
        return builder_block.build_navigation(configuration.value, page, request, context)


class NavigationPresenterBlockRegistry(BlockRegistry):

    # noinspection PyMethodMayBeStatic
    def should_register_block(self, app_label, local_identifier, block_type, block_args, block_kwargs, **kwargs):
        if not issubclass(block_type, NavigationPresenterMixin):
            raise RuntimeError("Registered block type must be a subclass of NavigationPresenterMixin")

        return True

    # noinspection PyMethodMayBeStatic
    def should_include_entry_for_container_block(self, identifier, entry, container_block):
        return True

    # noinspection PyMethodMayBeStatic
    def instantiate_block(self, identifier, entry, container_block):
        block_kwargs = dict(entry.block_kwargs)

        block = entry.block_type(*entry.block_args, **block_kwargs)
        block.set_name(identifier)

        return block


NAVIGATION_PRESENTATION_BLOCK_REGISTRY = NavigationPresenterBlockRegistry()
NAVIGATION_PRESENTATION_BLOCK_REGISTRY.define_procedures_in_caller_module("navigation_presenter")


class NavigationPresenterMixin:

    def render_navigation(self, configuration, root_node, render_location, request, context=None):
        return None


NavigationPresenterBlockValue = SwitchValue


class NavigationPresenterBlock(NavigationPresenterMixin, DynamicSwitchBlock):

    class Meta:
        presenter_blocks_function_name = APP_LABEL + ".blocks.navigation_presenter_block_choices"
        choice_label = "Select Navigation Presenter"

    def __init__(self, *args, **kwargs):

        presenter_blocks_function_name = kwargs.pop("presenter_blocks_function_name",
                                                    self._meta_class.presenter_blocks_function_name) # noqa

        super().__init__(*args,
                         presenter_blocks_function_name=presenter_blocks_function_name,
                         child_blocks_function_name=presenter_blocks_function_name,
                         **kwargs)

    def deconstruct(self):
        path, args, kwargs = blocks.Block.deconstruct(self)
        return path, args, kwargs

    def render_navigation(self, configuration, root_node, render_location, request, context=None):

        presenter_block = self.child_blocks[configuration.type]
        return presenter_block.render_navigation(configuration.value, root_node, render_location, request, context)


class PagePathBuilderBlock(NavigationBuilderMixin, blocks.StaticBlock):

    def build_navigation(self, configuration, page, request, context=None):

        node = None
        queryset = Page.objects.ancestor_of(page, inclusive=True).filter(depth__gt=1)

        for page in reversed(queryset):
            parent = NavigationNode.for_page(page, is_active=True)

            if node:
                parent.children.append(node)

            node = parent

        tmp_node = node
        depth = 0

        while tmp_node:
            tmp_node.depth = depth

            depth += 1

            if tmp_node.children:
                tmp_node = tmp_node.children[0]
            else:
                tmp_node = None

        return node


register_navigation_builder_block(APP_LABEL, "page_path", PagePathBuilderBlock, [], {}) # noqa


class PageTreeBuilderBlock(NavigationBuilderMixin, blocks.StructBlock):

    navigation_categories = DynamicMultipleChoiceBlock(
        choices_function_name=APP_LABEL + ".categories.navigation_category_choices"
    )

    def build_navigation(self, configuration, page, request, context=None):
        """
        Traverse all pages starting with the root page, gathering navigation links along the way
        :param page:
        :param request:
        :return: a navigation root node
        """

        navigation_categories = configuration.get('navigation_categories', [])
        root_page = root_page_for_request(request)
        active_page = page
        navigation = None

        stack = list([(root_page, 0, None, 0, None)])
        path = []

        while stack:

            page, child_index, children, index_in_parent, node = stack[-1] # noqa

            if children is None:
                # First visit
                children = select_page_children(page, navigation_categories)

                is_active = (active_page.url_path.startswith(page.url_path) if page and active_page else False)

                node = NavigationNode.for_page(page,
                                               is_active=is_active,
                                               depth=len(path),
                                               order=index_in_parent,
                                               path=list(path))

                if len(stack) > 1:
                    stack[-2][-1].children.append(node)
                else:
                    navigation = node

            if child_index >= len(children):
                # Last visit

                stack.pop()
                path.pop() if path else None

                continue

            stack[-1] = page, child_index + 1, children, index_in_parent, node

            state = children[child_index], 0, None, child_index, None
            stack.append(state)

            path.append(child_index)

        return navigation

register_navigation_builder_block(APP_LABEL, "page_tree", PageTreeBuilderBlock, [], {}) # noqa


class BeginNavigationLevelBlock(blocks.StaticBlock):

    class Meta:
        verbose_name = "Navigation Level Start"
        verbose_name_plural = "Navigation Level Starts"


class EndNavigationLevelBlock(blocks.StaticBlock):

    class Meta:
        verbose_name = "Navigation Level End"
        verbose_name_plural = "Navigation Level Ends"


InlineDefinitionBlockValue = blocks.StreamValue


class InlineDefinitionBlock(ModelPorterSupportMixin, NavigationBuilderMixin, blocks.StreamBlock):

    class Meta:
        label = "Navigation Definition"

    link = RichLinkBlock(label="Link", required=False,
                         user_editable_display_context=False,
                         user_editable_relationships=False)

    begin = BeginNavigationLevelBlock(label="Begin Level", required=False)
    end = EndNavigationLevelBlock(label="End Level", required=False)

    def deconstruct(self):
        return blocks.Block.deconstruct(self)

    # noinspection PyMethodMayBeStatic
    def from_repository(self, value, context):
        result = []

        for stream_item in value:

            block_type = stream_item["type"]
            value = stream_item["value"]

            item_block_def = self.child_blocks.get(block_type, None)

            if item_block_def is None:
                continue

            if isinstance(item_block_def, ModelPorterSupportMixin):
                value = item_block_def.from_repository(value, context)

            result.append({"type": block_type, "value": value})

        return result

    def build_navigation(self, configuration, page, request, context=None):

        path = [0]
        stack = [[]]

        for navigation_block in configuration:

            if navigation_block.block_type == 'begin':
                path.append(0)
                stack.append(stack[-1][-1].children)

            elif navigation_block.block_type == 'end':
                path.pop()
                stack.pop()

            elif navigation_block.block_type == 'link':

                node = navigation_block.value.as_navigation_node()
                node.depth = len(path)
                node.order = path[-1]
                node.path = list(path[1:])

                path[-1] = path[-1] + 1
                stack[-1].append(node)

        if len(stack[0]) != 1:

            root_page = WagtailFunctions.locate_page_for_path("/")
            navigation = NavigationNode(is_active=True, depth=0, order=0, path="/", page=root_page, rich_text="Home")
            navigation.children = stack[0]
        else:
            navigation = stack[0][0]

        return navigation


register_navigation_builder_block(APP_LABEL, "inline_definition", InlineDefinitionBlock, [], {}) # noqa


class TreeTemplatesBlock(NavigationPresenterMixin, blocks.StructBlock):

    class Meta:
        wrapper_template = APP_LABEL + '/blocks/navigation_wrapper.html'
        container_template = APP_LABEL + '/blocks/navigation_container.html'
        choices_template = APP_LABEL + '/blocks/navigation_choices.html'
        option_template = APP_LABEL + '/blocks/navigation_option.html'

        container_classname = ""

    NodeContext = namedtuple("NodeContext", ["navigation_id",
                                             "node",
                                             "node_url",
                                             "is_root",
                                             "render_location",
                                             "indentation",
                                             "options"])

    render_root = blocks.BooleanBlock(required=False, default=True)

    def deconstruct(self):
        return blocks.Block.deconstruct(self)

    # noinspection PyMethodMayBeStatic
    def create_node_context(self, navigation_id, request, node, render_location, indentation, options):

        node_url = node.determine_url(request)

        context = self.NodeContext(navigation_id=navigation_id,
                                   node=node, node_url=node_url, is_root=node and node.depth == 0,
                                   render_location=render_location, indentation=indentation,
                                   options=options)

        return context

    def contribute_to_options(self, configuration, root_node, render_location, request, options):
        pass

    # noinspection PyMethodMayBeStatic
    def render_element_template(self, template, node_context, did_enter, request):

        if not template:
            return ''

        context = node_context._asdict()  # noqa
        context["did_enter"] = did_enter
        context["did_exit"] = not did_enter

        output = render_to_string(template, context=context, request=request)
        return output

    # noinspection PyMethodMayBeStatic
    def render_wrapper(self, node_context, did_enter, request):

        template = node_context.options['wrapper_template']
        output = self.render_element_template(template, node_context, did_enter, request)
        return output

    # noinspection PyMethodMayBeStatic
    def render_container(self, node_context, did_enter, request):

        template = node_context.options['container_template']
        output = self.render_element_template(template, node_context, did_enter, request)
        return output

    # noinspection PyMethodMayBeStatic
    def render_choices(self, node_context, did_enter, request):

        template = node_context.options['choices_template']
        output = self.render_element_template(template, node_context, did_enter, request)
        return output

    # noinspection PyMethodMayBeStatic
    def render_option(self, node_context, did_enter, request):

        template = node_context.options['option_template']
        output = self.render_element_template(template, node_context, did_enter, request)
        return output

    def render_navigation(self, configuration, root_node, render_location, request, context=None):

        result = ''
        indent = 2
        indentation = ''

        scope = load_variable_scope(APP_LABEL, navigation_index=0)
        scope.navigation_index += 1
        navigation_id = 'navigation-{:d}'.format(scope.navigation_index)

        options = {key: getattr(self.meta, key) for key in dir(self.meta) if not key.startswith('_')}
        options.update(configuration)

        if context:
            options.update(context)

        self.contribute_to_options(configuration, root_node, render_location, request, options)

        render_root_option = options.get('render_root', self.child_blocks['render_root'].get_default())

        context_stack = []

        def visit(node, stack, enter):
            nonlocal result
            nonlocal indentation

            if enter:

                node_context = self.create_node_context(
                                    navigation_id, request, node, render_location, indentation, options)

                context_stack.append(node_context)

                result += self.render_container(node_context, enter, request)

                if node_context.is_root is False or render_root_option:
                    result += self.render_option(node_context, enter, request)

                if node.children:
                    result += self.render_choices(node_context, enter, request)

                indentation += ' ' * indent
            else:

                node_context = context_stack.pop()

                indentation = indentation[:-indent]

                if node.children:
                    result += self.render_choices(node_context, enter, request)

                result += self.render_option(node_context, enter, request)
                result += self.render_container(node_context, enter, request)

        wrapper_context = self.create_node_context(
                            navigation_id, request, root_node, render_location, indentation, options)

        result = self.render_wrapper(wrapper_context, True, request) + result

        root_node.traverse(visit)

        result = result + self.render_wrapper(wrapper_context, False, request)

        result = mark_safe(result)
        return result


class DropdownPresenterBlock(TreeTemplatesBlock):

    class Meta:
        js_pointer_initiator = APP_LABEL + ".PointerNavigation"
        js_touch_initiator = "adede.ui_components.HUD"
        classname = "dropdown"

        wrapper_template = APP_LABEL + '/blocks/dropdown/navigation_wrapper.html'
        container_template = APP_LABEL + '/blocks/dropdown/navigation_container.html'
        choices_template = APP_LABEL + '/blocks/dropdown/navigation_choices.html'
        option_template = APP_LABEL + '/blocks/dropdown/navigation_option.html'

    placement = blocks.ChoiceBlock(choices=[('placement-top-left', 'Top Left'),
                                            ('placement-top-right', 'Top Right'),
                                            ('placement-bottom-left', 'Bottom Left'),
                                            ('placement-bottom-right', 'Bottom Right')],

                                   required=True)

    margin = ComplexCSSUnitValueBlock()

    autohide = blocks.BooleanBlock(required=False, default=False)


register_navigation_presenter_block(APP_LABEL, "dropdown", DropdownPresenterBlock, [], {}) # noqa


class BreadcrumbsPresenterBlock(TreeTemplatesBlock):

    class Meta:
        classname = "breadcrumbs"

        wrapper_template = APP_LABEL + '/blocks/breadcrumbs/navigation_wrapper.html'
        container_template = APP_LABEL + '/blocks/breadcrumbs/navigation_container.html'
        choices_template = APP_LABEL + '/blocks/breadcrumbs/navigation_choices.html'
        option_template = APP_LABEL + '/blocks/breadcrumbs/navigation_option.html'

register_navigation_presenter_block(APP_LABEL, "breadcrumbs", BreadcrumbsPresenterBlock, [], {}) # noqa


class SiteMapPresenterBlock(TreeTemplatesBlock):

    class Meta:
        classname = "site_map"

        wrapper_template = APP_LABEL + '/blocks/site_map/navigation_wrapper.html'
        container_template = APP_LABEL + '/blocks/site_map/navigation_container.html'
        choices_template = APP_LABEL + '/blocks/site_map/navigation_choices.html'
        option_template = APP_LABEL + '/blocks/site_map/navigation_option.html'

register_navigation_presenter_block(APP_LABEL, "site_map", SiteMapPresenterBlock, [], {}) # noqa


SITE_NAVIGATION_PREFERENCE_TARGET = PreferenceTarget(APP_LABEL, 'site_navigation', 'Site Navigation')
BREADCRUMBS_PREFERENCE_TARGET = PreferenceTarget(APP_LABEL, 'breadcrumbs', 'Breadcrumbs Navigation')
SITE_MAP_PREFERENCE_TARGET = PreferenceTarget(APP_LABEL, 'site_map', 'Site Map Navigation')



# -*- coding: utf-8 -*-

import logging
import os

from django.db.models import DEFERRED
from django.core.files.storage import default_storage as DEFAULT_STORAGE # noqa
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe

from django.contrib.contenttypes.models import ContentType

from PIL import Image as PILImage

from .apps import get_app_label

__all__ = ['ChangeScope', 'ChangeScopeMixin']


APP_LABEL = get_app_label()


class ChangeScope(object):

    def __init__(self, navigation_style):
        self.navigation_style = navigation_style

    def __enter__(self):

        self.navigation_style.change_level += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self.navigation_style.change_level -= 1

        if self.navigation_style.change_count != 0 and self.navigation_style.change_level == 0:
            self.navigation_style.change_count = 0
            self.navigation_style.assemble()


class ChangeScopeMixin:

    class Meta:
        abstract = True

    @property
    def change_level(self):
        return self.change_level_

    @change_level.setter
    def change_level(self, value):
        self.change_level_ = value

    @property
    def change_count(self):
        return self.change_count_

    @change_count.setter
    def change_count(self, value):
        self.change_count_ = value

    @property
    def change_scope(self):
        return ChangeScope(self)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.change_level_ = 0
        self.change_count_ = 0

    def did_change(self):

        if self.change_level_ == 0:
            self.change_count_ = 0
            self.assemble()
        else:
            self.change_count_ += 1

    def assemble(self):
        pass




import re
import json
import copy
import secrets
from collections import namedtuple

from django.apps import apps
from django.db import models
from django.utils.safestring import mark_safe
from django.template.defaultfilters import capfirst

from django_auxiliaries.variable_scope import load_variable_scope

from .middleware import AnchorMiddleware
from .apps import get_app_label


__all__ = ['AnchorRegistry', 'Anchor', 'AnchorCategory', 'Token',
           'anchor_category_choices', 'anchor_category_choices_as_json',
           'cast_to_anchor_registry', 'current_anchor_registry_and_inventory',
           'number_format_choices', 'identifier_format_choices']

APP_LABEL = get_app_label()

middleware = AnchorMiddleware(None)


def anchor_category_choices():

    from .models import AnchorCategorySetting

    result = []
    settings = AnchorCategorySetting.objects.all()

    for setting in settings:
        result.append((setting.identifier, setting.name))

    return result


def anchor_category_choices_as_json():

    result = json.dumps(anchor_category_choices())
    return result


"""
def anchor_category_choices_as_dynamic_block_options():

    try:
        from wagtail_dynamic_blocks.schema import Option
    except ModuleNotFoundError:
        return []

    result = [Option(identifier=choice[0], label=choice[1]) for choice in anchor_category_choices()]

    for index, option in enumerate(result):
        option.value = option.identifier

    return result
"""


class NumberSymbol(object):

    @property
    def index(self):
        return self._index

    @property
    def text(self):
        return self._text

    def __init__(self):
        self._index = 0
        self._text = ''

    # noinspection PyMethodMayBeStatic
    def reset(self):
        return NumberSymbol()

    def increment(self, value=1):
        return self


class Base(NumberSymbol):

    @property
    def digits(self):
        return copy.deepcopy(self._digits)

    @property
    def period(self):
        return self._period

    @property
    def lowercase(self):
        return self._lowercase

    def __init__(self, digits=None, period=None, lowercase=False):
        super().__init__()

        if digits is None:
            # noinspection SpellCheckingInspection
            digits = list(zip(*'abcdefghijklmnopqrstuvwxyz'))

        for i in range(1, len(digits) - 1):
            if len(digits[i]) != len(digits[i - 1]):
                raise RuntimeError('Invalid digits specification for Base.')

        self._index = 0
        self._digits = digits
        self._period = period
        self._lowercase = lowercase

    # noinspection PyMethodMayBeStatic
    def reset(self):
        return Base(self._digits, self._period, self._lowercase)

    def increment(self, value=1):
        self._index += value

        value = self._index - 1

        if self._period:
            value = (value % self._period)

        value = value + 1

        self._text = ''

        p = len(self._digits[0])
        plane = 0

        while plane < len(self._digits):
            digits = self._digits[plane]
            self._text = digits[value % p] + self._text
            value //= p

            if value == 0:
                break

        if self._lowercase:
            self._text = self._text.lower()

        return self


DECIMAL = Base(digits=list(zip(*'1234567890')))
LETTER = Base(digits=list(zip(*'ABCDEFGHIJKLMNOPQRSTUVWXYZ')), lowercase=True)
ROMAN_NUMERAL = Base(digits=[['', 'I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX'],
                             ['', 'X', 'XX', 'XXX', 'XL', 'L', 'LX', 'LXX', 'LXXX', 'XC'],
                             ['', 'M', 'MM', 'MMM']
                            ], period=3999, lowercase=True)

"""
symbols = [Decimal Letter Numeral]
symbols = [Base(digits=[[i for i in range(10)]])
"""


def increment_order_state(state, level):
    while level > len(state):
        state.append(0)

    if level == len(state):
        state.append(-1)

    if level < len(state) - 1:
        state = state[:level + 1]

    state[level] += 1
    return state


def is_initial_state(state):
    for value in state:
        if value != 0:
            return False

    return True


class NumberFormat(object):

    def __init__(self, symbols=None, templates=None):
        super().__init__()

        if not symbols:
            symbols = [DECIMAL, DECIMAL, DECIMAL]

        if not templates:
            templates = ['{}', '{}', '({})']

        self._symbols = symbols
        self._templates = templates

    def __call__(self, state):

        result = ''

        for index, value in enumerate(state):

            symbol_index = index

            if symbol_index >= len(self._symbols):
                symbol_index = len(self._symbols) - 1

            symbol = self._symbols[symbol_index].reset()
            text = symbol.increment(value).text

            template_index = index

            if template_index >= len(self._templates):
                template_index = len(self._templates) - 1

            template = self._templates[template_index]
            text = template.format(text)
            result += text

        return result


NUMBER_FORMAT_REGISTRY = dict()


class NumberFormatRegistration:

    @property
    def identifier(self):
        return self.app_label + ":" + self.local_identifier

    def __init__(self, app_label, local_identifier, name, format_instance):
        self.app_label = app_label
        self.local_identifier = local_identifier
        self.name = name
        self.format_instance = format_instance

        app = apps.get_app_config(self.app_label)
        self.app_name = app.verbose_name


def register_number_format(app_label, local_identifier, name, format_instance):

    registration = NumberFormatRegistration(app_label=app_label, local_identifier=local_identifier,
                                            name=name, format_instance=format_instance)
    NUMBER_FORMAT_REGISTRY[registration.identifier] = registration
    return registration


def number_format_choices():

    result = [(identifier, number_format.name + " [{}]".format(number_format.app_name.title()))
              for identifier, number_format in NUMBER_FORMAT_REGISTRY.items()]

    return result


DEFAULT_NUMBER_FORMAT = register_number_format(
                            APP_LABEL, "numbers",
                            "Numbers 1, 1.1, 1.2, 2, ...",
                            NumberFormat(templates=['{}', '.{}']))

LETTERS_FORMAT = register_number_format(
                            APP_LABEL, "letters",
                            'Letters a, aa, ab, b, ...',
                            NumberFormat(symbols=[LETTER, LETTER, LETTER]))


class IdentifierFormat(object):

    def __init__(self):
        super().__init__()

    def __call__(self, category, anchor):

        if anchor is None or anchor.order is None:
            return ''

        result = "{}-".format(anchor.category)
        result += ".".join(["{:d}".format(level + 1) for level in anchor.order])
        return result


class LiteralIdentifier(IdentifierFormat):

    def __init__(self, category_prefix=False):
        super().__init__()
        self.__category_prefix = category_prefix

    def __call__(self, category, anchor):

        if anchor is None or anchor.order is None:
            return ''

        result = "{}-".format(anchor.category) if self.__category_prefix else ""
        result += anchor.identifier
        return result


IDENTIFIER_FORMAT_REGISTRY = dict()


class IdentifierFormatRegistration:

    @property
    def identifier(self):
        return self.app_label + ":" + self.local_identifier

    def __init__(self, app_label, local_identifier, name, format_instance):
        self.app_label = app_label
        self.local_identifier = local_identifier
        self.name = name
        self.format_instance = format_instance

        app = apps.get_app_config(self.app_label)
        self.app_name = app.verbose_name


def register_identifier_format(app_label, local_identifier, name, format_instance):

    registration = IdentifierFormatRegistration(app_label=app_label, local_identifier=local_identifier,
                                                name=name, format_instance=format_instance)

    IDENTIFIER_FORMAT_REGISTRY[registration.identifier] = registration
    return registration


def identifier_format_choices():

    result = [(identifier, identifier_format.name + " [{}]".format(identifier_format.app_name.title()))
              for identifier, identifier_format in IDENTIFIER_FORMAT_REGISTRY.items()]

    return result


DEFAULT_IDENTIFIER_FORMAT = register_identifier_format(
                            APP_LABEL, "generic",
                            "Generic [category]-1, [category]-1.1, [category]-1.2, [category]-2, ...",
                            IdentifierFormat())


USE_ANCHOR_IDENTIFIER_FORMAT = register_identifier_format(
                            APP_LABEL, "anchor_identifier",
                            "Anchor Identifier [anchor-id], ...",
                            LiteralIdentifier())


USE_PREFIXED_ANCHOR_IDENTIFIER_FORMAT = register_identifier_format(
                            APP_LABEL, "prefixed_anchor_identifier",
                            "Anchor Identifier [category]-[anchor-id], ...",
                            LiteralIdentifier(True))


class Anchor(object):

    @property
    def category(self):
        return self._category

    @property
    def identifier(self):
        return self._identifier

    @property
    def level(self):
        return self._level

    @property
    def order(self):
        return self._order

    @property
    def label_prefix(self):
        return self._label_prefix

    @property
    def label_prefix_order(self):
        return self._label_prefix_order

    @level.setter
    def level(self, value):
        self._level = value

    @order.setter
    def order(self, value):
        self._order = value

    @label_prefix.setter
    def label_prefix(self, value):
        self._label_prefix = value

    @label_prefix_order.setter
    def label_prefix_order(self, value):
        self._label_prefix_order = value

    def __init__(self, category, identifier, level=0, order=None):
        self._category = category
        self._identifier = identifier
        self._level = level
        self._order = order
        self._label_prefix = None
        self._label_prefix_order = None


class AnchorCategory(object):

    @property
    def identifier(self):
        return self._identifier

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def label_plural(self):
        return self._label_plural

    @label_plural.setter
    def label_plural(self, value):
        self._label_plural = value

    @property
    def label_short(self):
        return self._label_short

    @label_short.setter
    def label_short(self, value):
        self._label_short = value

    @property
    def link_css_class(self):
        return self._link_css_class

    @link_css_class.setter
    def link_css_class(self, value):
        self._link_css_class = value

    @property
    def label_css_class(self):
        return self._label_css_class

    @label_css_class.setter
    def label_css_class(self, value):
        self._label_css_class = value

    @property
    def number_format(self):
        return self._number_format

    @number_format.setter
    def number_format(self, value):
        self._number_format = value
        self._number_format_identifier = value.identifier if value else None

    @property
    def identifier_format(self):
        return self._identifier_format

    @identifier_format.setter
    def identifier_format(self, value):
        self._identifier_format = value
        self._identifier_format_identifier = value.identifier if value else None

    @property
    def sync_from_settings(self):
        return self._sync_from_settings

    def __init__(self, identifier, sync_from_settings=True):
        self._identifier = identifier
        self._sync_from_settings = sync_from_settings
        self._label = capfirst(self._identifier)
        self._label_plural = self._label + 's'
        self._label_short = self._label
        self._link_css_class = ''
        self._label_css_class = ''
        self._number_format_identifier = DEFAULT_NUMBER_FORMAT.identifier
        self._number_format = NUMBER_FORMAT_REGISTRY.get(self._number_format_identifier, DEFAULT_NUMBER_FORMAT).format_instance

        self._identifier_format_identifier = DEFAULT_IDENTIFIER_FORMAT.identifier
        self._identifier_format = IDENTIFIER_FORMAT_REGISTRY.get(self._identifier_format_identifier, DEFAULT_IDENTIFIER_FORMAT).format_instance

    def apply_settings(self):

        from .models import AnchorCategorySetting

        setting = AnchorCategorySetting.objects.get(identifier=self._identifier)

        if setting is None:
            return

        self._label = setting.label
        self._label_plural = setting.label_plural
        self._label_short = setting.label_short
        self._link_css_class = setting.link_css_class
        self._label_css_class = setting.label_css_class
        self._number_format_identifier = setting.number_format
        self._number_format = NUMBER_FORMAT_REGISTRY.get(self._number_format_identifier, DEFAULT_NUMBER_FORMAT).format_instance

        self._identifier_format_identifier = setting.identifier_format
        self._identifier_format = IDENTIFIER_FORMAT_REGISTRY.get(self._identifier_format_identifier, DEFAULT_IDENTIFIER_FORMAT).format_instance

    def create_anchor(self, identifier, level=0):
        anchor = Anchor(self.identifier, identifier, level)
        return anchor

    # noinspection PyMethodMayBeStatic
    def id_for_anchor(self, anchor):

        if anchor is None or anchor.order is None:
            return ''

        return self.identifier_format(self, anchor)

    # noinspection PyMethodMayBeStatic
    def url_for_anchor(self, anchor):

        if anchor is None:
            return ''

        return '#' + self.id_for_anchor(anchor)

    # noinspection PyMethodMayBeStatic
    def label_for_anchor(self, registry, inventory, anchor, label_format):

        if anchor is None or anchor.order is None:
            return ''

        if label_format not in ['label', 'short_label', 'no_prefix_label']:
            label_format = 'label'

        if label_format == 'short_label':
            category_prefix = self.label_short
        elif label_format == 'no_prefix_label':
            category_prefix = ''
        else:
            category_prefix = self.label

        if category_prefix:
            category_prefix += ' '

        if anchor.label_prefix:

            anchor_order = registry.get_anchor_order(inventory, do_create=False)

            if self.identifier in anchor_order:
                prefixes = anchor_order[self.identifier]
            else:
                prefixes = dict()

            if is_initial_state(prefixes.get(anchor.label_prefix, [0])):
                return category_prefix + anchor.label_prefix
            else:
                return category_prefix + anchor.label_prefix + self.number_format(anchor.label_prefix_order)

        return category_prefix + self.number_format(anchor.order)

    # noinspection PyMethodMayBeStatic
    def link_anchor(self, registry, anchor, inner):

        if self.link_css_class:
            attributes = ' class="' + self.link_css_class + '"'
        else:
            attributes = ''

        result = mark_safe('<a href="{}"{}>{}</a>'.format(registry.format_token('link', anchor), attributes, inner))
        return result

    # noinspection PyMethodMayBeStatic
    def label_anchor(self, registry, anchor, label_format):
        if not label_format:
            label_format = 'label'

        return registry.format_token(label_format, anchor)

    # noinspection PyMethodMayBeStatic
    def define_anchor(self, registry, anchor):
        result = registry.format_token('define', anchor)
        return result


Token = namedtuple('Token', ['start_offset', 'end_offset', 'directive', 'category', 'identifier', 'level'])

UNIQUE_ID_CATEGORY = ':unique_id'
COUNT_ANCHOR = ':count'
ANCHOR_ORDER = ':anchor_order'


class AnchorRegistry(models.Model):

    class Meta:
        abstract = True

    @property
    def anchor_categories(self):
        return list(self._anchor_categories.values())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name_in_context = 'anchor_inventory'
        self._anchor_categories = {}
        self._cached_inventory = None

    def reset(self):

        for anchor_category in self.anchor_categories:
            if anchor_category.sync_from_settings:
                anchor_category.apply_settings()

    def lookup_anchor_category(self, category, default=None, create=True):
        anchor_category = self._anchor_categories.get(category, None)

        if anchor_category is None:
            if create:
                anchor_category = AnchorCategory(category)
                anchor_category.apply_settings()
                self._anchor_categories[category] = anchor_category
            else:
                return default

        return anchor_category

    def add_anchor_category(self, category):
        self._anchor_categories[category.identifier] = category

    def remove_anchor_category(self, category):
        if self._anchor_categories[category.identifier] == category:
            del self._anchor_categories[category.identifier]

    def load_anchor_inventory(self, template_context):

        if template_context is not None:

            if self._name_in_context not in template_context:
                template_context[self._name_in_context] = {}

            self._cached_inventory = template_context[self._name_in_context]

        return self._cached_inventory

    # noinspection PyMethodMayBeStatic
    def count_anchors(self, inventory, category):

        if inventory is None:
            return 0

        category_dict = inventory.get(category, None) # noqa

        if category_dict is None:
            return 0

        return len(category_dict)

    def lookup_anchor(self, inventory, category, identifier, default=None, create=True):

        # inventory = self.load_anchor_inventory(template_context)

        if inventory is None:
            return default

        category_dict = inventory.get(category, None) # noqa

        if category_dict is None:
            if not create:
                return default

            category_dict = {}
            inventory[category] = category_dict

        anchor = category_dict.get(identifier, None)

        if anchor is None:
            if not create:
                return default

            anchor_category = self.lookup_anchor_category(category)
            anchor = anchor_category.create_anchor(identifier)
            category_dict[identifier] = anchor

        return anchor

    # noinspection PyMethodMayBeStatic
    def generate_unique_id(self, inventory):

        category_dict = inventory.get(UNIQUE_ID_CATEGORY, None) # noqa

        if category_dict is None:
            category_dict = {}
            inventory[UNIQUE_ID_CATEGORY] = category_dict

        while True:
            token = secrets.token_urlsafe(8)

            if token not in category_dict:
                category_dict[token] = None
                break

        return token

    # noinspection PyMethodMayBeStatic
    def count_category(self, inventory, category):

        category_dict = inventory.get(category, None) # noqa

        if category_dict is None:
            category_dict = {}
            inventory[category] = category_dict

        if COUNT_ANCHOR not in category_dict:
            category_dict[COUNT_ANCHOR] = 0

        result = category_dict[COUNT_ANCHOR] + 1
        category_dict[COUNT_ANCHOR] = result

        return result

    def url_for_anchor(self, inventory, category, identifier):
        anchor_category = self.lookup_anchor_category(category)
        anchor = self.lookup_anchor(inventory, category, identifier, create=False)
        if anchor is None:
            return ''
        result = anchor_category.url_for_anchor(anchor)
        return result

    def link_anchor(self, inventory, category, identifier, inner):
        anchor_category = self.lookup_anchor_category(category)
        anchor = self.lookup_anchor(inventory, category, identifier, create=True)
        result = anchor_category.link_anchor(self, anchor, inner)
        return result

    def label_anchor(self, inventory, category, identifier, label_format=None):
        anchor_category = self.lookup_anchor_category(category)
        anchor = self.lookup_anchor(inventory, category, identifier, create=True)
        result = anchor_category.label_anchor(self, anchor, label_format)
        return result

    def label_and_link_anchor(self, inventory, category, identifier, label_format=None):
        label_token = self.label_anchor(inventory, category, identifier, label_format)
        result = self.link_anchor(inventory, category, identifier, label_token)
        return result

    def define_anchor(self, inventory, category, identifier, level, label_prefix=None):
        anchor_category = self.lookup_anchor_category(category)
        anchor = self.lookup_anchor(inventory, category, identifier, create=True)
        anchor.level = level

        if label_prefix is not None:
            anchor.label_prefix = label_prefix

        result = anchor_category.define_anchor(self, anchor) # noqa
        return result

    # &bsol;def;results;figure;
    TOKEN_RE = re.compile(r'__anchor__:([^;]+);([^;]+);([^;]+);([0-9]+):')

    # noinspection PyMethodMayBeStatic
    def format_token(self, directive, anchor):
        return mark_safe('__anchor__:{};{};{};{:d}:'.format(
                    directive, anchor.category, anchor.identifier, anchor.level))

    # noinspection PyMethodMayBeStatic
    def match_next_token(self, text, pos=0):
        token_match = self.TOKEN_RE.search(text, pos=pos)

        if token_match is None:
            return None

        token = Token(start_offset=token_match.start(0),
                      end_offset=token_match.end(0),
                      directive=token_match.group(1),
                      category=token_match.group(2),
                      identifier=token_match.group(3),
                      level=int(token_match.group(4)))

        return token

    # noinspection PyMethodMayBeStatic
    def get_anchor_order(self, inventory, do_create=True):

        anchor_order = inventory.get(ANCHOR_ORDER, None) # noqa

        if anchor_order is None:

            if not do_create:
                return {}

            anchor_order = {}
            inventory[ANCHOR_ORDER] = anchor_order

        return anchor_order

    def order_anchor(self, inventory, category, identifier, level):

        anchor = self.lookup_anchor(inventory, category, identifier, create=True)
        anchor.level = level

        anchor_order = self.get_anchor_order(inventory)

        if category not in anchor_order:
            prefixes = {}
            anchor_order[category] = prefixes
        else:
            prefixes = anchor_order[category]

        state = self.assign_order_for_prefix(None, level, prefixes)
        anchor.order = list(state)

        if anchor.label_prefix:
            state = self.assign_order_for_prefix(anchor.label_prefix, level, prefixes)
            anchor.label_prefix_order = list(state)

    # noinspection PyMethodMayBeStatic
    def assign_order_for_prefix(self, anchor_prefix, level, prefixes):

        if anchor_prefix not in prefixes:
            state = []
            prefixes[anchor_prefix] = state
        else:
            state = prefixes[anchor_prefix]

        state = increment_order_state(state, level)
        prefixes[anchor_prefix] = state
        return state

    def resolve_token(self, inventory, directive, category, identifier):

        anchor = self.lookup_anchor(inventory, category, identifier, create=False)

        if anchor is None:
            return ''

        anchor_category = self.lookup_anchor_category(category)

        if directive == 'define':
            return anchor_category.id_for_anchor(anchor)
        elif directive == 'link':
            return anchor_category.url_for_anchor(anchor)
        elif directive.endswith('label'):
            return anchor_category.label_for_anchor(self, inventory, anchor, directive)

        return ''

    def serve(self, request):
        response = super().serve(request)  # noqa
        return middleware.process_template_response(request, response)

    def serve_preview(self, request, preview_mode):
        response = super().serve_preview(request, preview_mode)  # noqa
        return middleware.process_template_response(request, response)

    # noinspection PyMethodMayBeStatic
    def order_tokens(self, registry, inventory, tokens):

        for token in tokens:
            if token.directive != 'define':
                continue

            registry.order_anchor(inventory, token.category, token.identifier, token.level)

    # noinspection PyMethodMayBeStatic
    def apply_tokens(self, registry, inventory, tokens, text):

        result = ''
        pos = 0

        for token in tokens:

            if pos < token.start_offset:
                result += text[pos:token.start_offset]

            token_result = registry.resolve_token(inventory,
                                                  token.directive, token.category, token.identifier)

            result += token_result
            pos = token.end_offset

        if pos < len(text):
            result += text[pos:]

        return result

    # noinspection PyMethodMayBeStatic
    def parse_tokens(self, registry, text):

        tokens = []
        pos = 0

        while pos < len(text):

            token = registry.match_next_token(text, pos)

            if token is None:
                break

            tokens.append(token)
            pos = token.end_offset

        return tokens


def cast_to_anchor_registry(obj, default=None):

    if isinstance(obj, AnchorRegistry):
        return obj

    return default


def current_anchor_registry_and_inventory():

    current = load_variable_scope(APP_LABEL)
    return current.anchor_registry, current.anchor_inventory



from urllib.parse import quote

from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


__all__ = ['DISPLAY_CONTEXTS', 'LINK_RELATIONSHIPS', 'BaseLink', 'RichLink']


DISPLAY_CONTEXTS = [

    ("self", _("Same browsing context")),
    ("blank", _("New browsing context")),
    ("parent", _("Parent browsing context")),
    ("top", _("Top-level browsing context")),
]

LINK_RELATIONSHIPS = [

    ("alternate", _("Alternative")),
    ("author", _("Author")),
    ("bookmark", _("Bookmark")),
    ("external", _("External")),
    ("noopener", _("Deny access to current document from destination")), # noqa
    ("noreferrer", _("No reference to current document via Referer HTTP header")) # noqa
]


@deconstructible
class BaseLink:

    @property
    def rich_text(self):
        """
        A stylised, textual description of the link that appears in the context the link is embedded in.
        :return: A :py:class:`str` or :py:class:`wagtail.rich_text.RichText` instance.
        """

        return self.__rich_text

    @rich_text.setter
    def rich_text(self, value):
        self.__rich_text = value

    @property
    def display_context(self):
        """
        The display context for this link, one of {'self', 'blank', 'parent', 'top'}
        """
        return self.__display_context

    @display_context.setter
    def display_context(self, value):
        self.__display_context = value

    @property
    def display_context_as_html(self):
        """
        The HTML identifier corresponding to the :py:attr:`.display_context`.
        """
        return "_" + self.display_context

    @property
    def relationships(self):
        """
        A list of zero or more link relationships: {'alternate', 'author', 'bookmark', 'external', 'noopener', 'noreferrer'}
        """
        return self.__relationships[:]

    @relationships.setter
    def relationships(self, value):

        if value is None:
            value = []

        if len(value) == 1 and len(value[0]) == 0:
            value = []

        self.__relationships = list(value)

    @property
    def relationships_as_html(self):
        """
        A string suitable for embedding in HTML.
        """
        return " ".join(self.relationships)

    @property
    def media_item(self):
        return self.__media_item

    @media_item.setter
    def media_item(self, value):
        self.__media_item = value

    def __init__(self, *, rich_text=None, display_context=None, relationships=None, media_item=None):
        super().__init__()

        if not rich_text:
            rich_text = ''

        if display_context is None:
            display_context = 'self'

        if relationships is None:
            relationships = []

        self.__rich_text = rich_text
        self.__display_context = display_context
        self.__relationships = list(relationships) if relationships else []
        self.__media_item = media_item

    # noinspection PyMethodMayBeStatic
    def determine_url(self, request=None, current_site=None):
        return '#'


@deconstructible
class RichLink(BaseLink):

    @property
    def page(self):
        return self.__page

    @page.setter
    def page(self, value):
        self.__page = value

    @property
    def page_fragment(self):
        return self.__page_fragment

    @page_fragment.setter
    def page_fragment(self, value):
        self.__page_fragment = value

    @property
    def external_url(self):
        return self.__external_url

    @external_url.setter
    def external_url(self, value):
        self.__external_url = value

    def __init__(self, *, page=None, page_fragment="", external_url=None, **kwargs):

        super().__init__(**kwargs)

        self.__page = page
        self.__page_fragment = page_fragment
        self.__external_url = external_url

    def determine_url(self, request=None, current_site=None):

        if self.page:

            from .models import NavigationTarget

            page = self.page.specific
            result = None

            if isinstance(page, NavigationTarget):
                rich_link = page.rich_link

                if rich_link:
                    result = rich_link.as_rich_link().determine_url(request=request, current_site=current_site)

            else:
                result = page.get_url(request=request, current_site=current_site) # noqa

            if result is None:
                # page is not routable
                return "#"

            if self.page_fragment:
                result += '#' + quote(self.page_fragment) # noqa

            return result

        if self.page_fragment:
            return '#' + quote(self.page_fragment) # noqa

        return self.external_url

    @classmethod
    def for_page(cls, page, page_fragment=None, **kwargs):

        kwargs["page"] = page
        kwargs["page_fragment"] = page_fragment
        kwargs["rich_text"] = page.title

        result = cls(**kwargs)
        return result

    @classmethod
    def for_external_url(cls, external_url, **kwargs):

        kwargs["external_url"] = external_url

        result = cls(**kwargs)
        return result



from wagtail.admin.panels import FieldPanel

from .model_fields import NavigationCategoriesField

__all__ = 'chooses_navigation_categories'


def chooses_navigation_categories(cls):
    navigation_categories = NavigationCategoriesField(verbose_name="Navigation Categories")
    navigation_categories.contribute_to_class(cls, name="navigation_categories")

    panel = FieldPanel('navigation_categories')

    if 'content_panels' not in cls.__dict__:
        cls.content_panels = cls.content_panels + []

    cls.content_panels.extend([panel])

    return cls




# -*- coding: utf-8 -*-

from draftail_helpers.feature_registry import register_embed_feature

from wagtail.admin.rich_text.converters.html_to_contentstate import AtomicBlockEntityElementHandler, \
                                                                    InlineEntityElementHandler, Entity, KEEP_WHITESPACE
from wagtail.rich_text import EmbedHandler

from draftjs_exporter.dom import DOM

import re

from .anchor_feature_names import *
from .anchors import current_anchor_registry_and_inventory
from .apps import get_app_label

APP_LABEL = get_app_label()

"""

Anchor entity:

    action: ['define', 'link', 'label', 'labelled_link']
    category: 'figure'
    identifier: 'experiment_results'

"""

ACTION_PROPERTY = "action"
CATEGORY_PROPERTY = "category"
IDENTIFIER_PROPERTY = "identifier"
TEXT_PROPERTY = "text"
LEVEL_PROPERTY = "level"
EDITOR_SETTINGS_PROPERTY = "editor_settings"

HTML_ATTR_PREFIX = 'data-anchor-'


def editor_entity_data_to_html_attributes(props):

    attrs = {
        HTML_ATTR_PREFIX + EDITOR_SETTINGS_PROPERTY: props.get(EDITOR_SETTINGS_PROPERTY, ""),
    }

    return attrs


def html_attributes_to_editor_entity_data(attrs):

    props = {
        EDITOR_SETTINGS_PROPERTY: attrs.get(HTML_ATTR_PREFIX + EDITOR_SETTINGS_PROPERTY, "")
    }

    return props


def anchor_editor_entity_data_to_db(props, embed_type):

    attrs = editor_entity_data_to_html_attributes(props)

    attrs.update({
        'embedtype': embed_type,
        HTML_ATTR_PREFIX + ACTION_PROPERTY: props.get(ACTION_PROPERTY, ''),
        HTML_ATTR_PREFIX + CATEGORY_PROPERTY: props.get(CATEGORY_PROPERTY, ''),
        HTML_ATTR_PREFIX + IDENTIFIER_PROPERTY: props.get(IDENTIFIER_PROPERTY, ''),
        HTML_ATTR_PREFIX + TEXT_PROPERTY: props.get(TEXT_PROPERTY, ' '),
        HTML_ATTR_PREFIX + LEVEL_PROPERTY: props.get(LEVEL_PROPERTY, 0),
    })

    result = DOM.create_element("embed", attrs)  # *props['children'])
    return result


def anchor_db_to_editor_entity_data(attrs):

    props = html_attributes_to_editor_entity_data(attrs)

    props.update({
        ACTION_PROPERTY: attrs.get(HTML_ATTR_PREFIX + ACTION_PROPERTY, ''),
        CATEGORY_PROPERTY: attrs.get(HTML_ATTR_PREFIX + CATEGORY_PROPERTY, ''),
        IDENTIFIER_PROPERTY: attrs.get(HTML_ATTR_PREFIX + IDENTIFIER_PROPERTY, ''),
        TEXT_PROPERTY: attrs.get(HTML_ATTR_PREFIX + TEXT_PROPERTY, ' '),
        LEVEL_PROPERTY: attrs.get(HTML_ATTR_PREFIX + LEVEL_PROPERTY, 0),
    })

    return props


WHITESPACE_RE = re.compile("(\\s+)$", re.UNICODE)


def anchor_db_to_frontend_html(attrs):

    registry, inventory = current_anchor_registry_and_inventory()

    if not registry:
        return '[anchor undefined]'

    action = attrs.get(HTML_ATTR_PREFIX + ACTION_PROPERTY, '')
    category = attrs.get(HTML_ATTR_PREFIX + CATEGORY_PROPERTY, '')
    identifier = attrs.get(HTML_ATTR_PREFIX + IDENTIFIER_PROPERTY, '')
    level = attrs.get(HTML_ATTR_PREFIX + LEVEL_PROPERTY, 0)
    inner = attrs.get(HTML_ATTR_PREFIX + TEXT_PROPERTY, ' ')

    try:
        level = int(level)
    except ValueError:
        level = 0

    result = ''

    if action == 'define':
        result = registry.define_anchor(inventory, category, identifier, level)
    elif action == 'link':
        result = registry.link_anchor(inventory, category, identifier, inner) + " "
    elif action == 'label':
        result = registry.label_anchor(inventory, category, identifier) + " "
    elif action == 'labelled_link':
        result = registry.label_and_link_anchor(inventory, category, identifier) + " "

    return result


class InlineAnchorEmbedHandler(EmbedHandler):

    @staticmethod
    def get_model():
        raise NotImplementedError

    def __init__(self, feature_name):
        super().__init__()
        self.identifier = feature_name

    @classmethod
    def expand_db_attributes(cls, attrs):
        return anchor_db_to_frontend_html(attrs)


class InlineAnchorHandler(InlineEntityElementHandler):

    mutability = "IMMUTABLE"

    def __init__(self, entity_type):
        super().__init__(entity_type)
        self.inner_text = ' '

    def get_attribute_data(self, attrs):
        result = anchor_db_to_editor_entity_data(attrs)
        self.inner_text = result.get(TEXT_PROPERTY, self.inner_text)
        return result

    # noinspection SpellCheckingInspection
    def handle_endtag(self, name, state, contentstate):
        state.current_block.text += self.inner_text
        super().handle_endtag(name, state, contentstate)
        state.leading_whitespace = KEEP_WHITESPACE


def register_inline_anchor_define_feature(features):

    register_embed_feature(INLINE_ANCHOR_DEFINE_FEATURE_NAME,
                           INLINE_ANCHOR_DEFINE_ENTITY_TYPE,
                           'Define Anchor',
                           'Define anchor for selected text.',
                           InlineAnchorEmbedHandler(INLINE_ANCHOR_DEFINE_FEATURE_NAME),  # embed_handler
                           lambda props: anchor_editor_entity_data_to_db(props, INLINE_ANCHOR_DEFINE_FEATURE_NAME),  # entity_decorator
                           InlineAnchorHandler(INLINE_ANCHOR_DEFINE_ENTITY_TYPE),  # database_converter
                           features,
                           key_binding=None,
                           js_media_paths=[APP_LABEL + '/js/tour_guide.js'],
                           css_media_paths=[APP_LABEL + '/css/tour_guide.css']
                           )


def register_inline_anchor_link_feature(features):

    register_embed_feature(INLINE_ANCHOR_LINK_FEATURE_NAME,
                           INLINE_ANCHOR_LINK_ENTITY_TYPE,
                           'Link',
                           'Make the selection a reference to an element on this page.',
                           InlineAnchorEmbedHandler(INLINE_ANCHOR_LINK_FEATURE_NAME),  # embed_handler
                           lambda props: anchor_editor_entity_data_to_db(props, INLINE_ANCHOR_LINK_FEATURE_NAME),  # entity_decorator
                           InlineAnchorHandler(INLINE_ANCHOR_LINK_ENTITY_TYPE),  # database_converter
                           features,
                           key_binding=None,
                           js_media_paths=[APP_LABEL + '/js/tour_guide.js'],
                           css_media_paths=[APP_LABEL + '/css/tour_guide.css']
                           )


def register_inline_anchor_labelled_link_feature(features):

    register_embed_feature(INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME,
                           INLINE_ANCHOR_LABELLED_LINK_ENTITY_TYPE,
                           'Label Link',
                           'A labelled reference to an element on this page.',
                           InlineAnchorEmbedHandler(INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME),  # embed_handler
                           lambda props: anchor_editor_entity_data_to_db(props, INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME),  # entity_decorator
                           InlineAnchorHandler(INLINE_ANCHOR_LABELLED_LINK_ENTITY_TYPE),  # database_converter
                           features,
                           key_binding=None,
                           js_media_paths=[APP_LABEL + '/js/tour_guide.js'],
                           css_media_paths=[APP_LABEL + '/css/tour_guide.css']
                           )


feature_registrations = [register_inline_anchor_define_feature,
                         register_inline_anchor_link_feature,
                         register_inline_anchor_labelled_link_feature
                         ]

feature_names = [INLINE_ANCHOR_DEFINE_FEATURE_NAME,
                 INLINE_ANCHOR_LINK_FEATURE_NAME,
                 INLINE_ANCHOR_LABELLED_LINK_FEATURE_NAME
                 ]



from django.utils.deconstruct import deconstructible

from .links import RichLink

__all__ = ['NavigationNode']


@deconstructible
class NavigationNode(RichLink):

    def __init__(self, *, is_active=False, depth=0, order=0, path=None, **kwargs):
        super().__init__(**kwargs)

        self.is_active = is_active
        self.depth = depth
        self.order = order
        self.path = path if path is not None else list()
        self.children = []

    def traverse(self, callback):
        stack = [(self, 0)]

        while stack:
            node, index = stack[-1]  # noqa

            if index == 0:
                # node open
                callback(node, stack[:], True)

            if index >= len(node.children):
                # node close
                callback(node, stack[:], False)
                stack.pop()  # noqa
                continue

            stack[-1] = node, index + 1
            node = node.children[index]
            stack.append((node, 0))



from django.apps import apps
from .apps import get_app_label

__all__ = ['NavigationCategory', 'navigation_category_choices']

APP_LABEL = get_app_label()


class NavigationCategory:

    __REGISTRY = {}

    @property
    def identifier(self):
        return self.app_label + ":" + self.local_identifier

    def __init__(self, app_label, local_identifier, name):

        self.app_label = app_label
        self.local_identifier = local_identifier
        self.name = name

        app = apps.get_app_config(self.app_label)
        self.app_name = app.verbose_name
        self.__class__.__REGISTRY[self.identifier] = self

    @classmethod
    def all(cls):
        return list(cls.__REGISTRY.values())


MENU_NAVIGATION_CATEGORY = NavigationCategory(APP_LABEL, 'menu', 'Menu Navigation')
SITE_MAP_NAVIGATION_CATEGORY = NavigationCategory(APP_LABEL, 'site_map', 'Site Map Navigation')


def navigation_category_choices():

    return [(category.identifier, category.name) for category in NavigationCategory.all()]


"""
def navigation_category_options():

    from wagtail_dynamic_blocks.schema.controls import Option

    return [Option(identifier=choice[0], label=choice[1], value=choice[0]) for choice in navigation_category_choices()]
"""